package org.demo;

import java.util.HashMap;
import java.util.Map;
import java.io.*;

import org.apache.camel.CamelContext;
import org.apache.camel.Exchange;
import org.apache.camel.ExchangePattern;
import org.apache.camel.ProducerTemplate;
import org.apache.camel.builder.RouteBuilder;
import org.apache.camel.builder.AdviceWithRouteBuilder;
import org.apache.camel.component.mock.MockEndpoint;
import org.apache.camel.impl.DefaultExchange;
import org.apache.camel.test.spring.CamelSpringBootRunner;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;



import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

//import org.springframework.context.annotation.ImportResource;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.annotation.DirtiesContext.ClassMode;
import org.springframework.test.context.ActiveProfiles;

// import org.springframework.boot.test.web.client.TestRestTemplate;


@RunWith(CamelSpringBootRunner.class)
@SpringBootTest(    classes = org.demo.Application.class,
                    webEnvironment = SpringBootTest.WebEnvironment.DEFINED_PORT)
public class StreamTest extends Assert {

    private static final Logger LOG = LoggerFactory.getLogger(StreamTest.class);

    @Autowired
    private CamelContext context;

    @Autowired
    private ProducerTemplate template;

    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void baseBenchmarkClientIsHttp4_ServerIsSpringBootWithUndertow() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    .log("Client sending request stream...")
                    .bean(Helper.class, "prepareStream")
                    .to("http4:localhost:8080/camel/stream/benchmark?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "processStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching", 2147483664l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching", 2147483664l, response.getIn().getBody(Long.class).longValue());
    }



    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void smokeTestWithClientIsHttp4_ServerIsCamelNetty4Http() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    .log("Client sending request stream...")
                    .to("http4:localhost:10000/smoketest?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "processStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching.", 5l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching.", 5l, response.getIn().getBody(Long.class).longValue());
    }


    //TEST should cause broken pipe error
    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void testCamelNetty4HttpConsumerStreamUpload() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    .log("Client sending request stream...")
                    .bean(Helper.class, "prepareStream")
                    .to("http4:localhost:10000/upstream?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "processStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching.", 2147483664l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching.", 5l, response.getIn().getBody(Long.class).longValue());
    }



    //TEST should cause server to throw java.lang.OutOfMemoryError
    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void testCamelNetty4HttpConsumerStreamDownload() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    // .log("Client sending request stream...")
                    // .bean(Helper.class, "prepareStream")
                    .to("http4:localhost:10000/downstream?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "processStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching.", 5l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching.", 2147483664l, response.getIn().getBody(Long.class).longValue());
    }


    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void smokeTest_ClientIsCamelNetty4http_ServerIsCamelNetty4http() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    .log("Client sending request stream...")
                    .setHeader(Exchange.HTTP_METHOD, constant("POST")) 
                    .to("netty4-http:localhost:10000/smoketest?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "processStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching.", 5l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching.", 5l, response.getIn().getBody(Long.class).longValue());
    }

    //TEST should cause server to throw java.lang.OutOfMemoryError
    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void testCamelNetty4httpProducerStreamUpload() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    .log("Client sending request stream...")
                    .bean(Helper.class, "prepareStream")
                    .setHeader(Exchange.HTTP_METHOD, constant("POST"))
                    .to("netty4-http:localhost:8080/camel/stream/smoke?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "processStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");
        // request.setPattern(ExchangePattern.InOut);

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching.", 2147483664l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching.", 5l, response.getIn().getBody(Long.class).longValue());
    }

    //TEST should cause:
    //netty:    TooLongFrameException: Response entity too large
    //undertow: IOException: Broken pipe
    @Test
    @DirtiesContext(classMode = ClassMode.AFTER_EACH_TEST_METHOD)
    public void testCamelNetty4httpProducerStreamDownload() throws Exception {

        // We use a Camel route to ensure streaming is used properly
        // (Camel testing framework has issues handling streaming)
        context.addRoutes(new RouteBuilder() {
            @Override
            public void configure() throws Exception {

                from("direct:call")
                    .log("Client sending request stream...")
                    .setHeader(Exchange.HTTP_METHOD, constant("POST"))
                    .to("netty4-http://localhost:8080/camel/stream/benchmark?disableStreamCache=true")
                    .log("Client processing response stream...")
                    .bean(Helper.class, "asyncProcessStream")
                    .log("Client got response stream ${body} bytes long");
            }
        });

        //prepare new request
        DefaultExchange request = new DefaultExchange(context);
        request.getIn().setBody("dummy");
        // request.setPattern(ExchangePattern.InOut);

        //trigger request
        Exchange response = template.send("direct:call", request);

        //validate response success
        assertFalse("ups", response.isFailed());

        //validate request stream at server
        MockEndpoint mock = context.getEndpoint("mock:stream-size", MockEndpoint.class); 
        Long requestSize = mock.getExchanges().get(0).getIn().getBody(Long.class);
        assertEquals("request size not matching.", 5l, requestSize.longValue());

        //validate response stream at client
        assertEquals("response size not matching.", 2147483664l, response.getIn().getBody(Long.class).longValue());
    }

}
